#! fish -f
#
#  man -- reference manual page
#
#  Copyright (C) 1993-1994 Hitoshi Kihara.
#  This is free software; you can redistribute it and/or modify it under the
#  terms of the GNU General Public License, see the file COPYING.
#
#	$Id: man.fsh,v 1.1 1994/10/27 14:05:12 hkl Exp $
#

# set/reset shell variable
#
set   ciglob  nonomatch=drop
unset noalias noclobber noglob

# set defaults
#
set manver="man (script version) V1.11, Copyright (C) 1993-1994 Hitoshi Kihara"
set manp='A:/usr/man;A:/usr/local/man'
set secs='12345678ln'
set pager='less -s'
set nroff='cawf -man'
set ucomp='gzip -d'
set tempfile="man$$"
set dbname='whatis'
unalias rm;   alias rm 'rm -f'
unalias echo; alias echo '~~/echo -2'

# set interrupt routine ( jump this label )
onintr ctrl_c

# substitute by ENVIRONMENT variable
#
if ( $%?MANPATH ) set manp="$%{MANPATH:s.\\./.g.}"
if ( $%?MANSEC )  set secs="$%{MANSEC}"
if ( $%?PAGER )   set pager="$%{PAGER}"
if ( $%?NROFF )   set nroff="$%{NROFF}"
if ( $%?UCOMP )   set ucomp="$%{UCOMP}"
if ( $%?temp )    set tempfile="$%{temp:s./.\\.g.}\${tempfile}"

# initialize flags
#
set aproflag=false whatflag=false catflag=false

# check arguments
#
if ( "$0:t" == "apropos" ) set aproflag=true; goto term
if ( "$0:t" == "whatis"  ) set whatflag=true; goto term

while ( $#argv > 0 )
  switch ( $1 )
  case -f:
    set whatflag=true; shift; goto term
  case -k:
    set aproflag=true; shift; goto term
  case -V:
    echo $manver; shift
    breaksw
  case -M:
    shift
    if ( $#argv == 0 ) then
      echo '-M: missing path'; goto term
    endif
    set manp="${1:s.\\./.g}"; shift
    breaksw
  case -:
    set catflag=true; shift
    breaksw
  default:
    goto term
  endsw
end

if ( $#argv == 0 ) goto usage

term:
if ( $catflag == true ) set pager="cat"

# execute apropos
if ( $aproflag == true ) then
  if ( $#argv < 1 ) echo 'apropos what?'; goto terminate

  while ( $#argv > 0 )
    set find=false
    foreach mm ( ${manp:s/\;/ /g/} )
      if ( -f $mm/$dbname ) then
        egrep "$1" $mm/$dbname
        if ( $status != 1 ) set find=true
      else if ( ! -d $mm/$dbname ) then
        echo "$mm/${dbname}: No such file or directory"
      endif
    end
    shift
    if ( $find == false ) echo "${1}: nothing appropriate"
  end
  goto terminate
endif

# execute whatis
if ( $whatflag == true ) then
  if ( $#argv < 1 ) echo 'whatis what?'; goto terminate

  while ( $#argv > 0 )
    set find=false whatform="^([^ ]*, )*${1:t:r}(, [^ ]*)* \(.*\)[ 	]*- "
    foreach mm ( ${manp:s/\;/ /g/} )
      if ( -f $mm/$dbname ) then
        egrep -e "$whatform" $mm/$dbname
        if ( $status != 1 ) set find=true
      else if ( ! -d $mm/$dbname ) then
        echo "$mm/${dbname}: No such file or directory"
      endif
    end
    if ( $find == false ) echo "man: man entry for ${1:t:r} not found"
    shift
  end
  goto terminate
endif

# execute man
#
set mm="${manp:s/;/ /g/}"
set manp=''
foreach a ( $mm )
 set manp="${manp}${a:s.$.;.g.:s./;..g:s.;..g},"
end
set manp=${manp:s/$/;/g/:s/,;//g/}
set tempfile=${tempfile:s.\\\\.\\.g.}

set scanflag=false scan=$secs need=false find=false
while ( $#argv > 0 )
  if ( strlen "$1" == 1 && "$1" =~ [$secs] ) then
    set scanflag=true scan="$1" need=true
    shift; continue
  endif
  set scanflag=false

  # nroffed ( and compressed ) manuals
  set tt={$manp}/cat[$scan]/$1.[$scan]{,.[Zz],.gz}
  while ( $#tt > 0 )
    if ( ${tt[1]:t:r:al} == $1 ) then
      $pager ${tt[1]:s./.\\.g.}
    else ## if ( ${match:e} =~ Zz || ${match:e} == gz ) then
      echo -n 'Uncompressing page.  Wait...'
      $ucomp < $tt[1] > $tempfile
      echo ' done'
      $pager $tempfile
    endif
    set find=true; shift tt
  end

  # manuals ( roff style )
  if ( $find == false ) then
    set tt={$manp}/man[$scan]/$1.[$scan]{,.[Zz],.gz}
    while ( $#tt > 0 )
      if ( ${tt[1]:t:r:al} == $1 ) then
        echo -n 'Reformatting page.  Wait...'
        cat $tt[1] | eval $nroff > $tempfile
      else
        echo -n 'Uncompressing and Reformatting page.  Wait...'
        $ucomp < $tt[1] | eval $nroff > $tempfile
      endif
      echo ' done'
      $pager $tempfile
      set find=true; shift tt
    end
  endif

##
##    }jAt@Ckikg`Ήj
##
##    ꉞ͂Aɂ̓T|[gĂȂBieXgĂȂ ^^;j
##
##    TwentyOne ͕KvȂA LHA KvB
##
##    }jAfBNg README.DOC ɏ]č쐬ĂāA
##    Ⴆ cat1  foo.1 Ƃ}jA foo.Lzh ƂO
##    ɂɊi[Ă΁ALHA  p IvVgpă}jA
##    QƂ邱ƂłB
##    ɓɃ}jAt@C݂Ȃ̃`FbN͂ĂȂB
##
##    t@CʂɈkKv̂Ŗʓ|LB
##    ( }jAPɂɃt@CPKv )
##
##    man.x ɂ͂̋@\͖̂ŁAf TwentyOne  gzip g܂傤B
##
##---------i艺̍śus̐擪 #v폜j
#
#    if ( $find == false ) then
#      set tt={$manp}/cat[$scan]/${1}.Lzh
#      if ( x"$tt" != x ) then
#        if ( $%?PAGER == 0 ) setenv PAGER "$pager"
#        echo 'Invoking LHA.  Wait...'
#        lha p $tt[1]
#        set find=true
#      endif
#    endif
#
##---------i܂Łj
##

  # results
  if ( $find == false ) then
    if ( $need == true ) then
      echo "No entry for $1 in section $scan of the manual."
    else
      echo "No manual entry for ${1}."
    endif
  endif

  shift
end

# section number only?
#
if ( $scanflag == true && $need == true ) then
  echo "But what do you want from section ${scan}?"
  rm $tempfile; exit(1)
endif

# normal termination
#
terminate:
  rm $tempfile; exit(0)

# exceptions
#
usage:
  echo 'Usage:  man [-] [-M path] [ section ] name ...'
  echo '        man -k keyword ...'
  echo '        man -f file ...'
  exit(1)
ctrl_c:
  echo ' aborted (sorry)'; rm $tempfile; exit(1)

# eof
