#! fish -f
#
#  mkwhatis.fsh -- make 'whatis' database
#
#  Copyright (C) 1993-1994 Hitoshi Kihara.
#  This is free software; you can redistribute it and/or modify it under the
#  terms of the GNU General Public License, see the file COPYING.
#
#	$Id: mkwhatis.fsh,v 1.1 1994/10/27 14:05:26 hkl Exp $
#

# set/reset shell variable
#
set   ciglob  nonomatch=drop hugearg=indirect
unset noalias noclobber noglob

# set defaults
#
set mkwver="mkwhatis V1.11, Copyright (C) 1993-1994 Hitoshi Kihara"
set manp='A:/usr/man;A:/usr/local/man'
set secs='12345678ln'
set ucomp='gzip -d'
set awk='gawk'
set mkwdb="mkwhatis.db"
set dbname='whatis'
set tempfile="mkw$$"
unalias rm;   alias rm   'rm -f'
unalias echo; alias echo '~~/echo -2'

# set interrupt routine ( jump this label )
onintr exit

# substitute by ENVIRONMENT variable
#
if ( $%?MANPATH ) set manp="$%{MANPATH:s.\\./.g}"
if ( $%?MANSEC )  set secs="$%{MANSEC}"
if ( $%?UCOMP )   set ucomp="$%{UCOMP}"
if ( $%?AWK )     set awk="$%{AWK}"
if ( $%?temp )    set tempfile="$%{temp:s./.\\.g.}\${tempfile}"

# print version
echo $mkwver

# set script for AWK
set regform='
BEGIN			{s1="";s2=""}
/^#/ || $1 == ""	{next}
$1 == "ID1:"	{
	n=2;
	while($n!=""){
		if(s1==""){s1=sprintf("\\$1 == \\\"%s\\\"",$n)}
		else	{s1=s1 " || " sprintf("\\$1 == \\\"%s\\\"",$n)}
		n++
	}
}
$1 == "ID2:"	{
	n=2;
	while($n!=""){
		if(s2==""){s2=sprintf("\\$1 == \\\"%s\\\"",$n)}
		else	{s2=s2 " || " sprintf("\\$1 == \\\"%s\\\"",$n)}
		n++
	}
}
END		{
	printf("set regex1=(%s) regex2=(%s)",s1,s2);
}'
set mkwform1='#
BEGIN			{state=0}
NR > 20	{
	printf("%s: This looks like a normal text\n",fn)>"/dev/stderr";exit
}
$1 == ""		{next}
/^\.SH/			{sub("\.SH","")}
/^\./			{NR--;next}
/\\-/			{gsub("\\\\-","-")}
/.\b/			{gsub(".\b","")}'
set mkwform2='		{state=1;next}'
set mkwform3='		{if(state==2){print"";exit}}'
set mkwform4='
state == "1"	{
	n=1;$(NF+1)="-";
	while($n!="-"){fmt=fmt $n " ";n++}n++;$NF="";
	fmt=fmt ss;printf("%-22s  -",fmt);state=2
}
state == "2"	{
	buf="";
	while($n!=""){buf=buf " " $n;n++}
	printf("%s",buf);n=1
}'

# generate `mkwhatis' awk script
#
if ( "$0" != "$0:h" ) then
  set mkwdb="${0:h}/$mkwdb"
else if ( "$0:d" != "" ) then
  set mkwdb="${0:d}/$mkwdb"
endif
if ( ! -f $mkwdb ) then
  echo "mkwhatis: ${mkwdb}: No such file or directory"
  exit(1)
endif
set regform="${regform:s.	..g:s.
..g}"
eval `$awk -- "$regform" < ${mkwdb}`

rm $tempfile
~~/echo "$mkwform1"		>  $tempfile
~~/echo "${regex1}${mkwform2}"	>> $tempfile
~~/echo "${regex2}${mkwform3}"	>> $tempfile
~~/echo "$mkwform4"		>> $tempfile

# make `whatis' database
#
foreach mm ( ${manp:s/;/ /g/} )
  set whatis=$mm/$dbname
  rm $whatis
  set tt=( $mm/{cat,man}[$secs]/*.[$secs]{,.[Zz],.gz} )
  while ( $#tt > 0 )
    if ( "${tt[1]:t:ar}" == "${tt[1]:t:r}" ) then
      $awk -v "ss=(${tt[1]:e})" -v "fn=$tt[1]" \
          -f $tempfile  < ${tt[1]:s./.\\.g.} >> $whatis
    else
      $ucomp < $tt[1] | $awk -v "ss=(${tt[1]:r:e})" -v "fn=$tt[1]" \
          -f $tempfile  >> $whatis
    endif
    shift tt
  end
end

exit:
rm $tempfile
exit(0)

# eof
