/* lexer.c -- lexical analyzer for `mkwhatis.db'
 * Copyright (C) 1993-1994 Hitoshi Kihara.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef RCSID
static char rcsid[]="$Id: lexer.c,v 1.1 1994/10/27 14:06:07 hkl Exp $";
#endif /* RCSID */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "lexer.h"

/*------------------------------------*/
/* Local Functions                    */
/*------------------------------------*/
static int  tgetc(void);
static int  nextchar(void);

/*------------------------------------*/
/* Local Variables                    */
/*------------------------------------*/
#define TOKENLEN 512

static char * yyfilename;
static FILE * yyin;
static char yytext[TOKENLEN];
static int  yyeof;
static int  yylineno;
static int  lookaheadval;

/*----------------------------------------------*/
static int yylex()
{
  int    state;
  int    ch;
  char * p;

  if (yyeof == TRUE) return  DONE;

  state = 0;
  p     = yytext;
  ch    = nextchar();

  *p = 0;
  
  while (1) {
/*    printf("state = %d, ch = %d\n", state, ch); */
    switch (state) {
    case 0:
      if      (ch == '#')    { state = 4; }
#ifdef NODEF
      else if (ch == 0x0d)   { state = 3; }
#endif /* NODEF */
      else if (ch == '\n')   { tgetc(); yylineno++; return NL; }
      else if (ch == EOF)    { yyeof = TRUE; return DONE; }
      else if (!isblank(ch)) { state = 1; *p++ = ch; }
      tgetc(); ch = nextchar();
      break;
    case 1: /* ID */
      if      (isblank(ch)) { state = 2; }
#ifdef NODEF
      else if (ch == 0x0d)  { state = 3; }
#endif /* NODEF */
      else if (ch == '\n')  { state = 2; continue; }
      else if (ch == EOF)   { state = 2; yyeof = TRUE; }
      else                  { *p++ = ch; }
      tgetc(); ch = nextchar();
      break;
    case 2: /* end of ID */
      if (yytext[0] != 0) {
        *p = 0;
        if (strcmp(yytext, "ID1:") == 0) { return ID1; }
        if (strcmp(yytext, "ID2:") == 0) { return ID2; }
        return ID;
      }
      state = 0;
      break;
#ifdef NODEF
    case 3: /* LF */
      if (ch == 0x0a)   { tgetc(); ch = nextchar(); }
      state = 0;
      break;
#endif /* NODEF */
    case 4: /* comment */
#ifdef NODEF
      if      (ch == 0x0d)  { state = 3; }
#endif /* NODEF */
      if      (ch == '\n')  { yylineno++; state = 0; }
      else if (ch == EOF)   { yyeof = TRUE; return DONE; }
      tgetc(); ch = nextchar();
      break;
#ifdef DEBUG
    default:; printf("logical error! state=(%d)\n",state);
#endif /* DEBUG */
    }
  }
}

/*----------------------------------------------*/
static int tgetc()
{
  return  getc(yyin);
}

/*----------------------------------------------*/
static int nextchar()
{
  int c;

  if ((c = getc(yyin)) == EOF) return EOF;
  ungetc(c, yyin);
  return  c;
}

/*----------------------------------------------*/
/*  Interface Functions                         */
/*----------------------------------------------*/

/*----------------------------------------------*/
int lex_start_scan(char * name)
{
  if ((yyin = fopen(name, "r")) == NULL) return  FALSE;
  yyfilename = name;
  yylineno = 1;
  yyeof = FALSE;
  lex_next();

  return  TRUE;
}

/*----------------------------------------------*/
void lex_stop_scan()
{
  fclose(yyin);
}

/*----------------------------------------------*/
int lex_lookahead()
{
  return  lookaheadval;
}

/*----------------------------------------------*/
char * lex_get_text()
{
  return  yytext;
}

/*----------------------------------------------*/
void lex_next()
{
  lookaheadval = yylex();
}
/*----------------------------------------------*/
int lex_lineno()
{
  int lineno=yylineno;

  if (lookaheadval == NL) { lineno--; }
  return  lineno;
}


#ifdef D_LEXER
/*----------------------------------------------*/
/*  main() for debug                            */
/*----------------------------------------------*/
int main(int argc, char **argv)
{
  int ret;
  
  if (argc != 2) {
    fprintf(stderr, "usage: lexer file\n");
    exit(1);
  }
  ret = lex_start_scan(argv[1]);
  if (ret == FALSE) {
    fprintf(stderr, "Can't open %s\n", argv[1]);
    exit(1);
  }
  printf("enter loop\n");
  while (lex_lookahead() != DONE) {
    if (lex_lookahead() == NL) {
      printf(" :line(%d)\n", lex_lineno());
    }
    else {
      printf("token [%s] ", lex_get_text());
    }
    lex_next();
  }
  printf("exit loop\n");
  lex_stop_scan();
}
#endif /* D_LEXER */


/* eof */
