/* man -- reference manual page
 * Copyright (C) 1993-1994 Hitoshi Kihara.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef RCSID
static char rcsid[]="$Id: man.c,v 1.1 1994/10/27 14:06:33 hkl Exp $";
#endif /* RCSID */

#include "version.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#ifdef __HUMAN68K__
#define __DOS_INLINE__
#include <doslib.h>
#undef _filbuf		/* from <doslib.h> */
#else
#include <dirent.h>
#include <signal.h>
#include <unistd.h>

#endif /* __HUMAN68K__ */

/*------------------------------------------------------------------*/
/*  vg^Cv錾                                                */
/*------------------------------------------------------------------*/
extern          void man(char ** args);
extern          void apropos_whatis(int function, char ** args);
extern volatile void error_in_doscall(int code, char * mes);

#ifdef INTERNAL_GZIP
extern           int gzipmain (char * infile, char * outfile);
#endif /* INTERNAL_GZIP */

#ifdef DEBUG
extern          void dumpdtree(void);
#endif /* DEBUG */

/*------------------------------------------------------------------*/
/*  u[l                                                        */
/*------------------------------------------------------------------*/
#define TRUE	-1
#define FALSE	0

/*------------------------------------------------------------------*/
/* \̐錾у\bh`                                     */
/*------------------------------------------------------------------*/
#include "man.h"

char *          dirtbl[128];	/* man fBNge[u */
Directories_t *	dtree_top;	/* fBNgc[̃gbv */

#define Directory__scan(dp)    ( dp = dtree_top )
#define Directory__path(dp)    ( dirtbl[dp->index] )
#define Directory__section(dp) ( dp->sec )
#define Directory__next(dp)    ( dp = dp->next )
#define Directory__more(dp)    ( (int)dp )

#define File__scan(dp,fp,d)    ( fp = dp->file[d] )
#define File__name(fp)         ( fp->name )
#define File__extension(fp)    ( fp->ext )
#define File__compressed(fp)   ( fp->ext[0] )
#define File__next(fp)         ( fp = fp->next )
#define File__more(fp)         ( (int)fp )

/*------------------------------------------------------------------*/
/* bZ[W                                                       */
/*------------------------------------------------------------------*/
char usage_man[]="\
Usage:  man [-Vvis1ea] [-] [-M path] [section] name ...\x0d\x0a\
        man -k keyword ...\x0d\x0a\
        man -f file ...\x0d\x0a\
        man -L\x0d\x0a";
char usage_apropos[]="apropos what?\x0d\x0a";
char usage_whatis[]="whatis what?\x0d\x0a";

char nomem[]="man: not enough memory.\x0d\x0a";

char mkdtreemes[]="Making directory tree. wait... ";
char done[]="done\x0d\x0a";

#define ILLPOS 16
char illegal[]="Illegal option '?'\x0d\x0a";

char misspath[]="-M: missing path\x0d\x0a";

#define FORM_SEC 4
#if defined(__HUMAN68K__) || defined(MS_DOS)
char catform[]="\\cat?";
char manform[]="\\man?";
char tempname[]="\\man????1";
char wildname[]="\\*.*";
char delimitor[]="\\";
#else
char catform[]="/cat?";
char manform[]="/man?";
char tempname[]="//manXXXXX";
char delimitor[]="/";
#endif
char * formtbl[]= { manform, catform };


/* in man() */
char uncompmes[]="Uncompressing page. wait... ";
char reformmes[]="Reformatting page. wait... ";

char dot[]=".";

#define NOENTSEC 12
char noent1[]="No entry for ";
char noent2[]=" in section x of the manual.\x0d\x0a";
char nomanent[]="No manual entry for ";
char period[]=".\x0d\x0a";

char failtmp[]="e|t@C̍쐬Ɏs܂B\x0d\x0a";
#define FAILMESP 9
char failmes[]="\
ϐ ????? Ɏw肳ꂽR}hsł܂łB\x0d\x0a";

#define BUTWHATSEC 34
char butwhat[]="But what do you want from section x ?\x0d\x0a";

/*------------------------------------------------------------------*/
/* VOinh                                                 */
/*------------------------------------------------------------------*/
int onerror = FALSE;	/* G[nh CALL ꂽ TRUE */
int doneflg = FALSE;	/* PAGER ŝ TRUE ( abort Ȃ ) */
int save0= -1;		/* W(stdin) ێ */
int save1= -1;		/* Wo(stdout)ێ */
char tmpf[256];
char manf[256];

char abortmes[]="aborted (sorry)\x0d\x0a";

static void proc_ctrl_c()
{
#ifdef __HUMAN68K__
  struct _filbuf fbuf;

  if (FILES(&fbuf, manf, 32) >= 0) { DELETE(manf); }
  if (FILES(&fbuf, tmpf, 32) >= 0) { DELETE(tmpf); }
#else
  if (access(manf, F_OK)) { DELETE(manf); }
  if (access(tmpf, F_OK)) { DELETE(tmpf); }
#endif
  if (doneflg == FALSE) {
    if (save0 >= 0) { DUP2(save0, 0); CLOSE(save0); save0 = -1; }
    if (save1 >= 0) { DUP2(save1, 1); CLOSE(save1); save1 = -1; }
    FPUTS(abortmes, 2); onerror = TRUE;
  }
  EXIT2(1);
}

/*------------------------------------------------------------------*/
/* ftHgݒ                                                   */
/*------------------------------------------------------------------*/
#if defined(__HUMAN68K__) || defined(MS_DOS)
static const char Manpath[] = "A:/usr/man;A:/usr/local/man";
static const char Pager[]   = "less -s";
static const char Nroff[]   = "cawf -man";
static const char Temp[]    = "";
#else
static const char Manpath[] = "/usr/man:/usr/local/man";
static const char Pager[]   = "/usr/ucb/more -s";
static const char Nroff[]   = "/bin/nroff -man";
static const char Temp[]    = "/tmp";
#endif
static const char Mansecs[] = "12345678ln";
#ifndef INTERNAL_GZIP
static const char Ucomp[]   = "gzip -d";
#endif /* !INTERNAL_GZIP */
static const char Egrep[]   = "egrep -e";
static const char Cat[]     = "cat";

char manpath[256];
char mansecs[256];
char pager[256];
char nroff[256];
char ucomp[256];
char temp[256];
char egrep[256];

char progname[22];

int showver   = 0;	/* 0=hidden,      1=show version               */
int vb        = 0;	/* 0=verbose off, 1=verbose on                 */
int strcase   = 0;	/* 0=insensitive, 1=case sensitive             */
int scanall   = 1;	/* 0=scan 1st,    1=scan each,      2=scan all */
int showlic   = 0;	/* 0=hidden,      1=show license               */

/*------------------------------------------------------------------*/
/*  CC֐i񑀍j                                    */
/*------------------------------------------------------------------*/
static inline char * strcpy(char * dst, const char * src)
{
  while ((*dst++ = *src++) != 0);
}
static inline char * strncpy(char * dst, const char * src, size_t len)
{
  while (len-- > 0 && (*dst++ = *src++) != 0);
}
static inline char * strcat(char * dst, const char * src)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *src++) != 0);
}
static inline size_t strlen(const char * src)
{
  int i=-1;
  while (i++,*src++ != 0);
  return  i;
}
static inline int strcmp(const char * str1, const char * str2)
{
  int c1, c2;
  do {
    c1 = *str1++; c2 = *str2++;
  } while (c1 != 0 && c1 == c2);
  return  c1 - c2;
}
static inline int strcmpi(const char * str1, const char * str2)
{
  int c1, c2;
  do {
    c1 = *str1++; c2 = *str2++;
    if (/*isalpha(c1) &&*/ isupper(c1)) { c1 = tolower(c1); }
    if (/*isalpha(c2) &&*/ isupper(c2)) { c2 = tolower(c2); }
  } while (c1 != 0 && c1 == c2);
  return  c1 - c2;
}

static inline char * strcat2(
  char * dst,
  char * s1, char * s2
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0);
}

static inline char * strcat3(
  char * dst,
  char * s1, char * s2, char * s3
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0);
}

static inline char * strcat4(
  char * dst,
  char * s1, char * s2, char * s3, char * s4
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0); dst--;
  while ((*dst++ = *s4++) != 0);
}

static inline char * strcat5(
  char * dst,
  char * s1, char * s2, char * s3, char * s4, char * s5
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0); dst--;
  while ((*dst++ = *s4++) != 0); dst--;
  while ((*dst++ = *s5++) != 0);
}

/*------------------------------------------------------------------*/
/*  CC֐                                                  */
/*------------------------------------------------------------------*/
static inline int seperate(    /* t@C𕪊 */
  char *    src,
  char *    name,
  char *    sec,
  char *    ext
)
{
  *name = *sec = *ext = 0;
  /* '.' Ŏn܂t@C͂ */
  if (*src == '.') { return FALSE; }			/* drop ".foo" */

  /* t@Cu}jAvo */
  while (*src != '.') {
    /* gqȂ̂͂ */
    if ((*name++ = *src++) == 0) { return FALSE; }	/* drop "foo" */
  }
  *name = '\0'; src++;

  /* '.' ŏIĂ邩A'.'QĂ̂͂ */
  if (*src == '\0' || *src == '.') { return FALSE; }	/* drop "foo.",
  								"foo.." */

  /* t@CuZNVvo */
  while (*src != '.') {
    /* ZNVŏIĂ̂͂nj */
    if ((*sec++ = *src++) == 0) { return TRUE; }	/* ok "foo.1" */
  }
  *sec = '\0'; src++;

  /* t@Cuk̊gqvo */
  while (*src != '.') {
    /* k̊gqŏIĂ̂͂nj */
    if ((*ext++ = *src++) == 0) { return TRUE; }	/* ok "foo.1.gz" */
  }

  /* Ƃ͑S */
  return  FALSE;
}

char sectbl[256];
static inline void init_sectbl()/* ZNVԍe[u */
{
  char *msec = mansecs, *p=sectbl;
  int i;
  for (i=sizeof(sectbl); i > 0; *p++ = FALSE, i--);
  while (sectbl[*msec] = TRUE, *msec++ != 0);
}

static inline int isvalidsec(    /* LȃZNVԍ */
  char *   sec
)
{
  if (*sec == 0) return TRUE;
  else if (sec[1] != 0) return FALSE;
  else return  sectbl[*sec];
}
static inline int isvalidext(   /* LȊgq */
  char *   ext
)
{
  if (*ext == '\0') { return  TRUE; }	/* ext[0] */
  if (ext[1] == '\0') {
    if (*ext == 'Z' || *ext == 'z') { return  TRUE; }
    return  FALSE;
  }
    /* ext[0],         ext[1],           ext[2] */
  if (*ext++ == 'g' && *ext++ == 'z' && *ext == '\0') {
    return  TRUE;
  }
  return  FALSE;
}

static inline void basename(   /* pXt@C𓾂 */
  char *   dst,
  char *   src
)
{
  char *btm = src;

#if defined(__HUMAN68K__) || defined(MS_DOS)
  if (isalpha(src[0]) && src[1] == ':') { src++; btm = src++; }
#endif

  for ( ; *src != '\0'; src++) {
    if (*src == '/' || *src == '\\') { btm = src; }
  }
  if (*btm == '/' || *btm == '\\') { btm++; }
#if defined(__HUMAN68K__) || defined(MS_DOS)
  while (*btm != '.' && (*dst++ = *btm++) != 0);
  *dst = 0;
#else
  strcpy(dst, btm);
#endif
}

static inline void elimdelm(    /* A /  \ Pɂ */
  char *   dst,
  char *   src
)
{
  char prev = ' ';
  for ( ; *src != 0; src++ ) {
    if (*src == '/' || *src == '\\') {
      if (*src == prev) { continue; }
    }
    *dst++ = prev = *src;
  }
  /*if (dst[-1] == '/' || dst[-1] == '\\') { dst--; }*/
  if (prev == '/' || prev == '\\') { dst--; }
  *dst = '\0';
}

static inline void stdio_exec(  /* ẂCWo͂ɐڑĎs */
  int     handle_stdin,	  /* W͂ɐڑt@C̃nh */
  int     handle_stdout,  /* Wo͂ɐڑt@C̃nh */
  char *  prog,		  /* svO */
  char *  param,	  /* vÕp[^ */
  char *  startmes,	  /* Jnɏo͂郁bZ[W */
  char *  errormes	  /* ֘AϐiTŒj */
)
{
#ifdef __HUMAN68K__
  int ret;

  FPUTS(startmes, 2);
  DUP2(handle_stdin,  0); CLOSE(handle_stdin);
  DUP2(handle_stdout, 1); CLOSE(handle_stdout);
  ret = EXEC2(0, prog, param, 0);
  if (onerror != FALSE) { EXIT(); }
  DUP2(save0, 0); DUP2(save1, 1); FPUTS(done, 2);
  if (ret < 0) {
    strncpy(&failmes[FAILMESP], errormes, 5);
    error_in_doscall(ret, failmes);
  }
#endif /* __HUMAN68K__ */
}

/*------------------------------------------------------------------*/
/* fBNge[u                                         */
/*------------------------------------------------------------------*/
#if defined(__HUMAN68K__) || defined(MS_DOS)
#define SEP ';'
#else
#define SEP ':'
#endif
static inline void makedtree()
{
  char * top, *btm, **dst;

  top = btm = manpath; dst = dirtbl;
  while (*btm != '\0') {
    if (*btm == SEP) {
      *btm = '\0'; if (*top != '\0') { *dst++ = top; }
      while (*++btm == SEP); top = btm; continue;
    }
    btm++;
  }
  if (*top != '\0') { *dst++ = top; }
  *dst = NULL;
}

/*------------------------------------------------------------------*/
/* fBNgc[                                           */
/*------------------------------------------------------------------*/
#define MAXFILE 128
Files_t Files_1st[MAXFILE];

void allocdtree()
{
  char **               dirp;

  Directories_t *	dmy;
  Directories_t 	wkdir;
  Files_t *		dmy2;
  Files_t 		wkfile;

  Directories_t *	all_dirs;
  Files_t *		all_files = Files_1st;
  int			files_count = MAXFILE-1;

#ifdef __HUMAN68K__
  struct _filbuf fbuf;
  char * name=fbuf.name;
#else
  DIR * dirhandle;
  struct dirent * dp;
  char * name;
#endif /* __HUMAN68K__ */

  int i, dirnode;
  char path[256], *secno;
  char filename[22], sec[22], ext[22];
  int count;

  /* fBNgi[p̈m */
  for (i=0; dirtbl[i] != NULL; i++);
  all_dirs = (Directories_t *)malloc(sizeof(Directories_t)*i*strlen(mansecs));
  if (all_dirs == NULL) { FPUTS(nomem, 2); EXIT2(1); }

  dmy = &wkdir;

  for (dirp = dirtbl, count = 0;
       *dirp != NULL;
       dirp++, count++)					/* efBNgƂ */
  {
    for (secno = mansecs; *secno != '\0'; secno++)	/* eZNVƂ   */
    {
      formtbl[0][FORM_SEC] = formtbl[1][FORM_SEC] = *secno;
      dirnode = FALSE;
      for (i = 1; i >= 0; i--)		/* cat?, man? ̏ɃfBNg  */
      {
        elimdelm(path, *dirp);
#ifdef __HUMAN68K__
        strcat2(path, formtbl[i], wildname);
        if (FILES(&fbuf, path, 32) < 0) { continue; }
#else
        strcat(path, formtbl[i]);
        if ((dirhandle = opendir(path)) == NULL) { continue; }
        if ((dp = readdir(dp)) == NULL) { closedir(dirhandle); continue; }
#endif /* __HUMAN68K__ */

	if (dirnode == FALSE) {
	  dmy = dmy->next = all_dirs++;
	  dmy->index = count;
	  dmy->sec[0] = *secno; dmy->sec[1] = '\0';
	  dmy->file[0] = dmy->file[1] = (Files_t *)dmy->next = NULL;
	  dirnode = TRUE;
	}
        dmy2 = &wkfile;

	do {				/* fBNg̊et@CƂ    */
#ifndef __HUMAN68K__
          name = dp->d_name;
#endif /* !__HUMAN68K__ */
          if (seperate(name, filename, sec, ext) != FALSE &&
              sec[0] == *secno && sec[1] == 0 && isvalidext(ext))
          {
            if (files_count-- < 0) {
              all_files = (Files_t *)malloc(sizeof(Files_t) * MAXFILE);
              if (all_files == NULL) { FPUTS(nomem, 2); EXIT2(1); }
              files_count = MAXFILE-1;
            }
	    dmy2 = dmy2->next = all_files++;
	    strcpy(dmy2->name, filename);
	    strcpy(dmy2->ext, ext);
	    dmy2->next = NULL;
          }
#ifdef __HUMAN68K__
        } while (NFILES(&fbuf) >= 0);
#else
        } while ((dp = readdir(dirhandle)) != NULL);
        closedir(dirhandle);
#endif /* __HUMAN68K__ */
        dmy->file[i] = wkfile.next;
      }
    }
  }

  dtree_top = wkdir.next;
}

/*------------------------------------------------------------------*/
/*  main(): C֐                                              */
/*------------------------------------------------------------------*/
int main(int argc, char **argv)
{
  /****************************/
  /* @\tO               */
  /****************************/
  int functions = 0;	/* 0:man, 1:apropos, 2:whatis */
  /****************************/
  char *p;
  /**/

  save0 = DUP(0); if (save0 < 0) { error_in_doscall(save0, NULL); }
  save1 = DUP(1); if (save1 < 0) { error_in_doscall(save1, NULL); }

#ifdef __HUMAN68K__
  INTVCS(0xfff1, proc_ctrl_c);		/* $FFF1:CTRL-C ABORT xN^ */
  INTVCS(0xfff2, proc_ctrl_c);		/* $FFF2:ERROR ABORT xN^  */
#else
  signal(SIGINT, proc_ctrl_c);
#endif /* __HUMAN68K__ */

  manf[0] = tmpf[0] = 0;

  /****************************/
  /* ϐ`FbN         */
  /****************************/
  if (GETENV("MANOPT",  0, manpath)  >= 0) {
    for (p = manpath; *p != 0; p++) {
      if (*p == 'V') { showver=1; continue; } /* show version     */
      if (*p == 'v') { vb = 1;    continue; } /* verbose          */
      if (*p == 'i') { strcase=0; continue; } /* case insensitive */
      if (*p == 's') { strcase=1; continue; } /* case sensitive   */
      if (*p == '1') { scanall=0; continue; } /* scan 1st  match  */
      if (*p == 'e') { scanall=1; continue; } /* scan each match  */
      if (*p == 'a') { scanall=2; continue; } /* scan all  match  */
    }
  }
  if (GETENV("MANPATH", 0, manpath) < 0) { strcpy(manpath, Manpath); }
  if (GETENV("MANSEC",  0, mansecs) < 0) { strcpy(mansecs, Mansecs); }
  if (GETENV("PAGER",   0, pager)   < 0) { strcpy(pager,   Pager);   }
  if (GETENV("NROFF",   0, nroff)   < 0) { strcpy(nroff,   Nroff);   }
  if (GETENV("UCOMP",   0, ucomp)   < 0) {
#ifdef INTERNAL_GZIP
   ucomp[0] = 0;
#else
   strcpy(ucomp,   Ucomp);
#endif /* INTERNAL_GZIP */
  }
  if (GETENV("temp",    0, temp)    < 0) { strcpy(temp,    Temp);    }
  if (GETENV("EGREP",   0, egrep)   < 0) { strcpy(egrep,   Egrep);   }

  init_sectbl();

  /****************************/
  /* R}h`FbN       */
  /****************************/
  basename(progname, argv[0]);
  if      (strcmpi(progname, "apropos") == 0) { functions = 1; }
  else if (strcmpi(progname, "whatis")  == 0) { functions = 2; }

  /****************************/
  /* `FbN             */
  /****************************/
  argv++;
  while (*argv != NULL)
  {
    if (argv[0][0] != '-')  { break; }
    if (argv[0][1] == '\0') { strcpy(pager, Cat); argv++; continue; }
    if (argv[0][2] == '\0') {
      if (argv[0][1] == 'L') { showlic=1; break; } /* show license   */
      if (argv[0][1] == '-') { argv++;    break; } /* no more option */
      if (argv[0][1] == 'M') {
        argv++;
        if (*argv == NULL) { FPUTS(misspath, 2); EXIT2(1); }
        strcpy(manpath, *argv++); continue;
      }
      if (functions == 0) {
        if (argv[0][1] == 'k') { functions = 1; argv++; break; } /* apropos */
        if (argv[0][1] == 'f') { functions = 2; argv++; break; } /* whatis  */
      }
    }
    if (functions != 0) { break; }
    for (p = &argv[0][1]; *p != 0; p++) {
      if (*p == 'V') { showver=1; continue; } /* show version     */
      if (*p == 'v') { vb = 1;    continue; } /* verbose          */
      if (*p == 'i') { strcase=0; continue; } /* case insensitive */
      if (*p == 's') { strcase=1; continue; } /* case sensitive   */
      if (*p == '1') { scanall=0; continue; } /* scan 1st  match  */
      if (*p == 'e') { scanall=1; continue; } /* scan each match  */
      if (*p == 'a') { scanall=2; continue; } /* scan all  match  */
      /* sȃIvV */
      illegal[ILLPOS] = *p; FPUTS(illegal, 2); EXIT2(1);
    }
    argv++;
  }

  /* license \ */
  if (showlic != 0) {
   FPUTS(version, 2);
#ifdef INTERNAL_GZIP
   FPUTS(gzipcode, 2);
#endif /* INTERNAL_GZIP */
   FPUTS(license, 2);
   EXIT2(1);
  }

  /* o[W\ */
  if (showver != 0) { FPUTS(version, 2); }

  /* gp@\ďI */
  if (*argv == 0) {
    char *usage[]={ usage_man, usage_apropos, usage_whatis };
    FPUTS(usage[functions], 2);
    EXIT2(1);
  }

  /* fBNge[u쐬 */
  makedtree();

  /* @\ɂ菈I */
  if (functions == 0) { /* man */
    if (vb != 0) { FPUTS(mkdtreemes, 2); }
    allocdtree();
    if (vb != 0) { FPUTS(done, 2); }
#ifdef DEBUG
    if (vb != 0) { dumpdtree(); }
#endif /* DEBUG */
    man(argv);
  }
  else {
    apropos_whatis(functions, argv);
  }

  return  0;
}

/*------------------------------------------------------------------*/
/*  man(): man ֐                                             */
/*------------------------------------------------------------------*/
void man(char ** args)
{
  Directories_t *	Dhd;
  Files_t	*	Fhd;

  char pagerprm[256], * pagerend = pagerprm;
  char ucompprm[256];
  char nroffprm[256];
  char      prm[256];
  char * out, * sec, *this_sec;
  int fno0, fno1;
  int ret;
  int form;             /* 1...cat, 0...man */
  int need;             /* TRUE if no manual name with section  */
  int find;             /* TRUE if found manual in cat? or man? */

  int dopager=FALSE;
  int doucomp=FALSE;
  int donroff=FALSE;

  elimdelm(tmpf, temp); strcat(tmpf, tempname);

  out = sec = NULL; need = FALSE;
  while (*args != NULL) {
    /* ZNVw肳ꂽꍇAZNVԍo^ */
    if (isvalidsec(*args) != FALSE) {
      sec = *args++; need = TRUE; continue;
    }
    out = NULL;
    for (Directory__scan(Dhd);		/* fBNǧJn   */
         Directory__more(Dhd);		/* fBNg݂ԁA   */
         Directory__next(Dhd)		/* XƃfBNg */
        ){
      /* ZNVԍ`FbNiZNVw肳ꂽꍇ̂݁j */
      this_sec = Directory__section(Dhd);
      if (sec != NULL && *sec != *this_sec) { continue; }

      find = FALSE;

      for (form = 1; form >= 0; form--){/* cat?, man? ̏Ɍ    */
        for (File__scan(Dhd, Fhd, form);/* t@ČJn   */
             File__more(Fhd);		/* t@C݂ԁA   */
             File__next(Fhd)		/* Xƃt@C */
            ){
          if (strcase == 0) {	/* case insensitive */
            if (strcmpi(File__name(Fhd), *args) != 0) { continue; }
          }
          else {		/* case sensitive */
            if (strcmp(File__name(Fhd), *args)  != 0) { continue; }
          }
	  find = TRUE;		/* }jA */

	  strcpy(manf, tmpf);

	  formtbl[form][FORM_SEC] = *this_sec;
	  elimdelm(prm, Directory__path(Dhd));
	  strcat5(prm, formtbl[form], "\\", File__name(Fhd), dot, this_sec);
				/* delimitor */
	  out = prm;
     /****************************/
     /* CALL 'UNCOMPESS'         */
     /****************************/
	  if (File__compressed(Fhd) != FALSE) {	/* is compressed */
	    out = manf;
	    if (form == 0) {		/* is pre-roff */
	      out = tmpf;
	    }
	    strcat2(prm, dot, File__extension(Fhd));
#ifdef INTERNAL_GZIP
            if (ucomp[0] == 0) {
              FPUTS(uncompmes, 2);
              gzipmain( prm, out );
              FPUTS(done, 2);
            }
            else
#endif /* INTERNAL_GZIP */
            {
	      if (doucomp == FALSE) {
	        if ((ret = EXEC2(2, ucomp, ucompprm, 0)) < 0) {
		  strncpy(&failmes[FAILMESP], "UCOMP", 5);
		  error_in_doscall(ret, failmes);
	        }
	        doucomp = TRUE;
	      }
	      if ((fno0=   OPEN(prm, 0)) < 0){error_in_doscall(fno0,NULL);}
	      if ((fno1=MAKETMP(out,32)) < 0){error_in_doscall(fno1,failtmp);}
              stdio_exec(fno0, fno1, ucomp, ucompprm, uncompmes, "UCOMP");
	    }
	  }
     /****************************/
     /* CALL 'NROFF'             */
     /****************************/
	  if (form == 0) {		/* is pre-roff */
	    if (donroff == FALSE) {
	      if ((ret = EXEC2(2, nroff, nroffprm, 0)) < 0) {
		strncpy(&failmes[FAILMESP], "NROFF", 5);
		error_in_doscall(ret, failmes);
	      }
	      donroff = TRUE;
	    }
	    if ((fno0=   OPEN(out,  0)) < 0){error_in_doscall(fno0,NULL);}
	    if ((fno1=MAKETMP(manf,32)) < 0){error_in_doscall(fno1,failtmp);}
            stdio_exec(fno0, fno1, nroff, nroffprm, reformmes, "NROFF");
            out = manf; DELETE(tmpf);
	  }
     /****************************/
     /* CALL 'PAGER'             */
     /****************************/
	  if (dopager == FALSE) {
	    if ((ret = EXEC2(2, pager, pagerprm, 0)) < 0) {
	      strncpy(&failmes[FAILMESP], "PAGER", 5);
	      error_in_doscall(ret, failmes);
	    }
	    for (/*pagerend = pagerprm*/; *pagerend != 0; pagerend++);
	    *pagerend++ = ' '; dopager = TRUE;
	  }
	  strcpy(pagerend, out);
	  pagerprm[0] = (unsigned char)strlen(&pagerprm[1]);
	  doneflg = TRUE;
	  if ((ret = EXEC2(0, pager, pagerprm, 0)) < 0) {
	    strncpy(&failmes[FAILMESP], "PAGER", 5);
	    error_in_doscall(ret, failmes);
	  }
	  doneflg = FALSE;
	  if (onerror != FALSE) { EXIT(); }

	  DELETE(manf);
	}
	if (find != FALSE) break;
      }
      if (find != FALSE) {
        if (scanall == 0) return; /* ŏɌ}jÂݕ\     */
        if (scanall == 1) break;  /* }jAPɂPt@C\ */
      }
    }
    if (out == NULL) {
      if (sec != NULL) {
	noent2[NOENTSEC] = *sec;
	FPUTS(noent1, 2); FPUTS(*args, 2); FPUTS(noent2, 2);
      }
      else {
	FPUTS(nomanent, 2); FPUTS(*args, 2); FPUTS(period, 2);
      }
    }
    need = FALSE;
    args++;
  }
  if (sec != NULL && need != FALSE) {
    butwhat[BUTWHATSEC] = *sec;
    FPUTS(butwhat, 2);
  }
}

/*------------------------------------------------------------------*/
/*  apropos_whatis(): apropos, whatis ֐                      */
/*------------------------------------------------------------------*/
#if defined(__HUMAN68K) || defined(MS_DOS)
char filewhatis[]="\\whatis";	/* delimitor */
#else
char filewhatis[]="/whatis";	/* delimitor */
unsigned char *newargv[5]={ "/bin/egrep", "-e", 0, 0, 0 };
#endif

char noegrep[]="man: egrep not found\x0d\x0a";
char noappmes[]=": nothing appropriate\x0d\x0a";

char whatform1[]="^([^ ]*, )*";
char whatform2[]="(, [^ ]*)* \\(.*\\)[ 	]*- ";
char nowhat[]="man: man entry for ";
char notfound[]=" not found\x0d\x0a";
char nosuch[]=": no such file or directory\x0d\x0a";

void apropos_whatis(int function, char ** args)
{
#ifdef __HUMAN68K__
  struct _filbuf fbuf;
#endif /* __HUMAN68K__ */
  char egrepprm[256], form[256], *p1;
  char prm[256], path[256];
  char ** dirp;
  int  find;
  int  ret;
#if !defined(__HUMAN68K__) && !defined(MS_DOS)
  int pid;
#endif

#ifdef __HUMAN68K__
  if ((ret = EXEC2(2, egrep, egrepprm, 0)) < 0) {
    strncpy(&failmes[FAILMESP], "EGREP", 5);
    error_in_doscall(ret, failmes);
  }
#endif /* __HUMAN68K__ */

  function--;

  for ( ; *args != NULL; args++ ) {
    find = FALSE;
    if (function == 0) {
      p1 = *args;
    }
    else {
      strcpy(form, whatform1); strcat2(form, *args, whatform2);
      p1 = form;
    }
    for (dirp = dirtbl; *dirp != NULL; dirp++) {
      elimdelm(path, *dirp); strcat(path, filewhatis);
#ifdef __HUMAN68K__
      if (FILES(&fbuf, path, 48) < 0) {
        FPUTS(path, 2); FPUTS(nosuch, 2); continue;
      }
      strcpy(prm, egrepprm); strcat4(prm, " \"", p1, "\" ", path);
      prm[0] = (unsigned char)strlen(&prm[1]);
      ret = EXEC2(0, egrep, prm, 0);
#else
      if (!access(path, F_OK)) {
        FPUTS(path, 2); FPUTS(nosuch, 2); continue;
      }
# ifdef MS_DOS
      ret = spawnl(P_WAIT, egrep, egrep, &egrepprm[1], p1, path, NULL);
# else
      new_argv[2] = p1; newargv[3] = path;
      if ((pid = fork()) == 0) {
        execv(new_argv[0], new_argv);
        fprintf(stderr, "Couldn't exec the program %s\n", new_argv[0]);
        _exit(1);
      }
      if (wait(&ret) != pid) {
      	fprintf(stderr, "process fork failed, status=%d\n", ret);
        _exit(1);
      }
# endif /* MS_DOS */
#endif /* __HUMAN68K__ */
      if      (ret == 0) { find = TRUE; }
      else if (ret <  0) {
        strncpy(&failmes[FAILMESP], "EGREP", 5);
        error_in_doscall(ret, failmes);
      }
    }
    if (find == FALSE) {
      if (function == 0) {
                          FPUTS(*args, 2); FPUTS(noappmes, 2);
      }
      else {
        FPUTS(nowhat, 2); FPUTS(*args, 2); FPUTS(notfound, 2);
      }
    }
  }
}

/*------------------------------------------------------------------*/
/*  error_in_doscall(): DOS CALL G[p֐                   */
/*------------------------------------------------------------------*/
static char * errormessage[] = {
0, 0, 
"w肵t@C͌܂B\x0d\x0a",                 /* 2 */
"w肵fBNg͌܂B\x0d\x0a",               /* 3 */
"I[vĂt@C܂B\x0d\x0a",               /* 4 */
"fBNg{[x̓ANZXł܂B\x0d\x0a", /* 5 */
0, 0,
"ŝɕKvȃ܂B\x0d\x0a",               /* 8 */
0, 0, 0, 0,
"t@C̎wɌ肪܂B\x0d\x0a",                   /*13 */
0, 0, 0, 0, 0,
"̃t@C͏݂ł܂B\x0d\x0a",                 /*19 */
0, 0, 0,
"fBXNςŃt@C܂B\x0d\x0a",           /*23 */
"fBNgςŃt@C܂B\x0d\x0a"        /*24 */
/* 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0 */
};

static char errormessage80[] =
"t@C݂Ă܂B\x0d\x0a";


volatile void error_in_doscall(int code, char * mes)
{
#ifdef __HUMAN68K__
  struct _filbuf fbuf;
  char * mesp;

  code = -code;
  if (code == 80) {
    FPUTS(errormessage80, 2);
  }
  else if (code <= 24) {
    if ((mesp = errormessage[code]) != NULL) { FPUTS(mesp, 2); }
#ifdef DEBUG
    else { fprintf(stderr, "code=(%d)\n", code); }
#endif /* DEBUG */
  }
  if (mes != NULL) { FPUTS(mes, 2); }
  if (FILES(&fbuf, manf, 32) >= 0) { DELETE(manf); }
  if (FILES(&fbuf, tmpf, 32) >= 0) { DELETE(tmpf); }
#else
  perror(progname);
#endif
  EXIT2(1);
}

#ifdef DEBUG
/*------------------------------------------------------------------*/
/*  dumpdtree(): fobOp֐ifBNgc[\j     */
/*------------------------------------------------------------------*/
void dumpdtree()
{
  Directories_t *	Dhd;
  Files_t	*	Fhd;

  char path[256], *sec;
  int form;

  for (Directory__scan(Dhd);
       Directory__more(Dhd);
       Directory__next(Dhd)) {

    for (form = 1; form >= 0; form--) {

      sec = Directory__section(Dhd);
      formtbl[form][FORM_SEC] = *sec;
      elimdelm(path, Directory__path(Dhd); strcat(path, formtbl[form]);

      printf("Directory: [%s]\n", path);
    
      for (File__scan(Dhd, Fhd, form);
           File__more(Fhd);
           File__next(Fhd)) {

	if (File__compressed(Fhd) != FALSE) {
          sprintf(path, "%s.%s.%s", File__name(Fhd), sec, File__extension(Fhd));
	}
	else {
          sprintf(path, "%s.%s", File__name(Fhd), sec);
	}
	printf("  filename: [%s]\n", path);
      }
    }
  }
}
#endif /* DEBUG */


/* eof */
