/* mkwhatis -- make 'whatis' database
 * Copyright (C) 1993-1994 Hitoshi Kihara.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef RCSID
static char rcsid[]="$Id: mkwhatis.c,v 1.1 1994/10/27 14:07:04 hkl Exp $";
#endif /* RCSID */

#include "mkwver.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#ifdef __HUMAN68K__
#define __DOS_INLINE__
#include <doslib.h>
#undef _filbuf		/* from <doslib.h> */
#endif /* __HUMAN68K__ */

/*------------------------------------------------------------------*/
/*  vg^Cv錾                                                */
/*------------------------------------------------------------------*/
#include "lexer.h"	/* from lexer.c  */
#include "parser.h"	/* from parser.c */

extern          void gen_mkwhatis(void);
extern          void make_whatis(void);
extern volatile void error_in_doscall(int code, char * mes);

#ifdef INTERNAL_GZIP
extern           int gzipmain (char * infile, char * outfile);
#endif /* INTERNAL_GZIP */

#ifdef DEBUG
extern          void dumpdtree(void);
#endif /* DEBUG */

/*------------------------------------------------------------------*/
/*  u[l                                                        */
/*------------------------------------------------------------------*/
#define TRUE	-1
#define FALSE	0

/*------------------------------------------------------------------*/
/* \̐錾у\bh`                                     */
/*------------------------------------------------------------------*/
#include "man.h"

char *          dirtbl[128];	/* man fBNge[u */
Directories_t *	dtree_top;	/* fBNgc[̃gbv */

#define Directory__scan(dp)    ( dp = dtree_top )
#define Directory__path(dp)    ( dirtbl[dp->index] )
#define Directory__index(dp)   ( dp->index )
#define Directory__section(dp) ( dp->sec )
#define Directory__next(dp)    ( dp = dp->next )
#define Directory__more(dp)    ( (int)dp )

#define File__scan(dp,fp,d)    ( fp = dp->file[d] )
#define File__name(fp)         ( fp->name )
#define File__extension(fp)    ( fp->ext )
#define File__compressed(fp)   ( fp->ext[0] )
#define File__next(fp)         ( fp = fp->next )
#define File__more(fp)         ( (int)fp )

/*------------------------------------------------------------------*/
/* bZ[W                                                       */
/*------------------------------------------------------------------*/
char usage_mkwhatis[]="\
Usage:  mkwhatis [-Vv] [-M path]\x0d\x0a";

char nomem[]="mkwhatis: not enough memory.\x0d\x0a";

char nodb[]="mkwhatis: not found 'mkwhatis.db'.\x0d\x0a";

char prsmkwdb[]=  "Reading 'mkwhatis.db'. wait... ";
char mkdtreemes[]="Making directory tree. wait... ";
char genmkwhat[]= "Generating awk script. wait... ";
char done[]="done\x0d\x0a";

#define ILLPOS 16
char illegal[]="Illegal option '?'\x0d\x0a";

char misspath[]="-M: missing path\x0d\x0a";

#define FORM_SEC 4
char catform[]="\\cat?";	    /* delimitor */
char manform[]="\\man?";	    /* delimitor */
char * formtbl[]= { manform, catform };


/* in make_whatis() */
char uncompmes[]="Uncompressing page. wait... ";

char dot[]=".";
char period[]=".\x0d\x0a";

char failtmp[]="e|t@C̍쐬Ɏs܂B\x0d\x0a";
#define FAILMESP 9
char failmes[]="\
ϐ ????? Ɏw肳ꂽR}hsł܂łB\x0d\x0a";

/*------------------------------------------------------------------*/
/* VOinh                                                 */
/*------------------------------------------------------------------*/
int onerror = FALSE;	/* G[nh CALL ꂽ TRUE */
int doneflg = FALSE;	/* PAGER ŝ TRUE ( abort Ȃ ) */
int save0= -1;		/* W(stdin) ێ */
int save1= -1;		/* Wo(stdout)ێ */
char awkscript[256];
char tmp1[256];

char abortmes[]="aborted (sorry)\x0d\x0a";

static void proc_ctrl_c()
{
  struct _filbuf fbuf;

  if (FILES(&fbuf, awkscript, 32) >= 0) { DELETE(awkscript); }
  if (FILES(&fbuf, tmp1,      32) >= 0) { DELETE(tmp1); }
  if (doneflg == FALSE) {
    if (save0 >= 0) { DUP2(save0, 0); CLOSE(save0); save0 = -1; }
    if (save1 >= 0) { DUP2(save1, 1); CLOSE(save1); save1 = -1; }
    FPUTS(abortmes, 2); onerror = TRUE;
  }
  EXIT2(1);
}

/*------------------------------------------------------------------*/
/* ftHgݒ                                                   */
/*------------------------------------------------------------------*/
const char Manpath[] = "A:/usr/man;A:/usr/local/man";
const char Mansecs[] = "12345678ln";
#ifndef INTERNAL_GZIP
const char Ucomp[]   = "gzip -d";
#endif /* !INTERNAL_GZIP */
const char Temp[]    = "";
const char Awk[]     = "gawk";

char manpath[256];
char mansecs[256];
char ucomp[256];
char temp[256];
char awk[256];

char progname[22];

char tempname[]="\\mkw????1";		/* delimitor */
char wildname[]="\\*.*";		/* delimitor */
char file_mkwhatis[]="\\mkwhatis.db";	/* delimitor */
char file_whatis[]="\\whatis";		/* delimitor */

char mkwhatis_db[256];

int showver   = 0;	/* 0=hidden,      1=show version               */
int vb        = 0;	/* 0=verbose off, 1=verbose on                 */
int showlic   = 0;	/* 0=hidden,      1=show license               */

/*------------------------------------------------------------------*/
/*  CC֐i񑀍j                                    */
/*------------------------------------------------------------------*/
static inline char * strcpy(char * dst, const char * src)
{
  while ((*dst++ = *src++) != 0);
}
static inline char * strncpy(char * dst, const char * src, size_t len)
{
  while (len-- > 0 && (*dst++ = *src++) != 0);
}
static inline char * strcat(char * dst, const char * src)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *src++) != 0);
}
static inline size_t strlen(const char * src)
{
  int i=-1;
  while (i++,*src++ != 0);
  return  i;
}
static inline int strcmp(const char * str1, const char * str2)
{
  int c1, c2;
  do {
    c1 = *str1++; c2 = *str2++;
  } while (c1 != 0 && c1 == c2);
  return  c1 - c2;
}
static inline int strcmpi(const char * str1, const char * str2)
{
  int c1, c2;
  do {
    c1 = *str1++; c2 = *str2++;
    if (/*isalpha(c1) &&*/ isupper(c1)) { c1 = tolower(c1); }
    if (/*isalpha(c2) &&*/ isupper(c2)) { c2 = tolower(c2); }
  } while (c1 != 0 && c1 == c2);
  return  c1 - c2;
}

static inline char * strcat2(
  char * dst,
  char * s1, char * s2
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0);
}

static inline char * strcat3(
  char * dst,
  char * s1, char * s2, char * s3
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0);
}

static inline char * strcat4(
  char * dst,
  char * s1, char * s2, char * s3, char * s4
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0); dst--;
  while ((*dst++ = *s4++) != 0);
}

static inline char * strcat5(
  char * dst,
  char * s1, char * s2, char * s3, char * s4, char * s5
)
{
  while (*dst++ != 0 ); dst--;
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0); dst--;
  while ((*dst++ = *s3++) != 0); dst--;
  while ((*dst++ = *s4++) != 0); dst--;
  while ((*dst++ = *s5++) != 0);
}

static inline char * strcpycat(
  char * dst, char * s1, char * s2
)
{
  while ((*dst++ = *s1++) != 0); dst--;
  while ((*dst++ = *s2++) != 0);
}

/*------------------------------------------------------------------*/
/*  CC֐                                                  */
/*------------------------------------------------------------------*/
static inline int seperate(    /* t@C𕪊 */
  char *    src,
  char *    name,
  char *    sec,
  char *    ext
)
{
  *name = *sec = *ext = 0;
  /* '.' Ŏn܂t@C͂ */
  if (*src == '.') { return FALSE; }			/* drop ".foo" */

  /* t@Cu}jAvo */
  while (*src != '.') {
    /* gqȂ̂͂ */
    if ((*name++ = *src++) == 0) { return FALSE; }	/* drop "foo" */
  }
  *name = '\0'; src++;

  /* '.' ŏIĂ邩A'.'QĂ̂͂ */
  if (*src == '\0' || *src == '.') { return FALSE; }	/* drop "foo.",
  								"foo.." */

  /* t@CuZNVvo */
  while (*src != '.') {
    /* ZNVŏIĂ̂͂nj */
    if ((*sec++ = *src++) == 0) { return TRUE; }	/* ok "foo.1" */
  }
  *sec = '\0'; src++;

  /* t@Cuk̊gqvo */
  while (*src != '.') {
    /* k̊gqŏIĂ̂͂nj */
    if ((*ext++ = *src++) == 0) { return TRUE; }	/* ok "foo.1.gz" */
  }

  /* Ƃ͑S */
  return  FALSE;
}

static inline int isvalidext(   /* LȊgq */
  char *   ext
)
{
  if (*ext == '\0') { return  TRUE; }	/* ext[0] */
  if (ext[1] == '\0') {
    if (*ext == 'Z' || *ext == 'z') { return  TRUE; }
    return  FALSE;
  }
    /* ext[0],         ext[1],           ext[2] */
  if (*ext++ == 'g' && *ext++ == 'z' && *ext == '\0') {
    return  TRUE;
  }
  return  FALSE;
}

static inline void dirname(   /* pXfBNg𓾂 */
  char *   dst,
  char *   src
)
{
  char *btm = dst;
  int flag = FALSE;

#if defined(__HUMAN68K__) || defined(MS_DOS)
  if (isalpha(src[0]) && src[1] == ':') {
    *dst++ = *src++; *dst++ = *src++; btm = dst;
  }
#endif
  for ( ; *src != '\0'; *dst++ = *src++) {
    if (*src == '/' || *src == '\\') { btm = dst; flag = TRUE; }
  }
  if (flag == FALSE) { *btm++ = '.'; }
  *btm = 0;
}

static inline void basename(   /* pXt@C𓾂 */
  char *   dst,
  char *   src
)
{
  char *btm = src;

#if defined(__HUMAN68K__) || defined(MS_DOS)
  if (isalpha(src[0]) && src[1] == ':') { src++; btm = src++; }
#endif

  for ( ; *src != '\0'; src++) {
    if (*src == '/' || *src == '\\') { btm = src; }
  }
  if (*btm == '/' || *btm == '\\') { btm++; }
#if defined(__HUMAN68K__) || defined(MS_DOS)
  while (*btm != '.' && (*dst++ = *btm++) != 0);
  *dst = 0;
#else
  strcpy(dst, btm);
#endif
}

static inline void elimdelm(    /* A /  \ Pɂ */
  char *   dst,
  char *   src
)
{
  char prev = ' ';
  for ( ; *src != 0; src++ ) {
    if (*src == '/' || *src == '\\') {
      if (*src == prev) { continue; }
    }
    *dst++ = prev = *src;
  }
  /*if (dst[-1] == '/' || dst[-1] == '\\') { dst--; }*/
  if (prev == '/' || prev == '\\') { dst--; }
  *dst = '\0';
}

static inline void stdio_exec2(  /* ẂCWo͂ɐڑĎs */
  int     handle_stdin,	  /* W͂ɐڑt@C̃nh */
  int     handle_stdout,  /* Wo͂ɐڑt@C̃nh */
  char *  prog,		  /* svO */
  char *  param,	  /* vÕp[^ */
  char *  errormes	  /* ֘AϐiTŒj */
)
{
  int ret;

  DUP2(handle_stdin,  0); DUP2(handle_stdout, 1);
  ret = EXEC2(0, prog, param, 0);
  if (onerror != FALSE) { EXIT(); }
  DUP2(save0, 0); DUP2(save1, 1);
  if (ret < 0) {
    strncpy(&failmes[FAILMESP], errormes, 5);
    error_in_doscall(ret, failmes);
  }
}

/*------------------------------------------------------------------*/
/* fBNge[u                                         */
/*------------------------------------------------------------------*/
static inline void makedtree()
{
  char * top, *btm, **dst;

  top = btm = manpath; dst = dirtbl;
  while (*btm != '\0') {
    if (*btm == ';') {
      *btm = '\0'; if (*top != '\0') { *dst++ = top; }
      while (*++btm == ';'); top = btm; continue;
    }
    btm++;
  }
  if (*top != '\0') { *dst++ = top; }
  *dst = NULL;
}

/*------------------------------------------------------------------*/
/* fBNgc[                                           */
/*------------------------------------------------------------------*/
#define MAXFILE 128
Files_t Files_1st[MAXFILE];

void allocdtree()
{
  char **               dirp;

  Directories_t *	dmy;
  Directories_t 	wkdir;
  Files_t *		dmy2;
  Files_t 		wkfile;

  Directories_t *	all_dirs;
  Files_t *		all_files = Files_1st;
  int			files_count = MAXFILE-1;

  struct _filbuf fbuf;
  int i, dirnode;
  char path[256], *secno;
  char filename[22], sec[22], ext[22];
  int count;

  /* fBNgi[p̈m */
  for (i=0; dirtbl[i] != NULL; i++);
  all_dirs = (Directories_t *)malloc(sizeof(Directories_t)*i*strlen(mansecs));
  if (all_dirs == NULL) { FPUTS(nomem, 2); EXIT2(1); }

  dmy = &wkdir;

  for (dirp = dirtbl, count = 0;
       *dirp != NULL;
       dirp++, count++)					/* efBNgƂ */
  {
    for (secno = mansecs; *secno != '\0'; secno++)	/* eZNVƂ   */
    {
      formtbl[0][FORM_SEC] = formtbl[1][FORM_SEC] = *secno;
      dirnode = FALSE;
      for (i = 1; i >= 0; i--)		/* cat?, man? ̏ɃfBNg  */
      {
        elimdelm(path, *dirp); strcat2(path, formtbl[i], wildname);

        if (FILES(&fbuf, path, 32) < 0) { continue; }

	if (dirnode == FALSE) {
	  dmy = dmy->next = all_dirs++;
	  dmy->index = count;
	  dmy->sec[0] = *secno; dmy->sec[1] = '\0';
	  dmy->file[0] = dmy->file[1] = (Files_t *)dmy->next = NULL;
	  dirnode = TRUE;
	}
        dmy2 = &wkfile;

	do {				/* fBNg̊et@CƂ    */
          if (seperate(fbuf.name, filename, sec, ext) != FALSE &&
              sec[0] == *secno && sec[1] == 0 && isvalidext(ext))
          {
            if (files_count-- < 0) {
              all_files = (Files_t *)malloc(sizeof(Files_t) * MAXFILE);
              if (all_files == NULL) { FPUTS(nomem, 2); EXIT2(1); }
              files_count = MAXFILE-1;
            }
	    dmy2 = dmy2->next = all_files++;
	    strcpy(dmy2->name, filename);
	    strcpy(dmy2->ext, ext);
	    dmy2->next = NULL;
          }
        } while (NFILES(&fbuf) >= 0);

        dmy->file[i] = wkfile.next;
      }
    }
  }

  dtree_top = wkdir.next;
}

/*------------------------------------------------------------------*/
/*  main(): C֐                                              */
/*------------------------------------------------------------------*/
int main(int argc, char **argv)
{
  struct _filbuf fbuf;

  char *p;

  awkscript[0] = tmp1[0] = 0;

  /****************************/
  /* ẂCWo͂ۑ */
  /****************************/
  save0 = DUP(0); if (save0 < 0) { error_in_doscall(save0, NULL); }
  save1 = DUP(1); if (save1 < 0) { error_in_doscall(save1, NULL); }

  /****************************/
  /* VOinhݒ     */
  /****************************/
#ifdef __HUMAN68K__
  INTVCS(0xfff1, proc_ctrl_c);		/* $FFF1:CTRL-C ABORT xN^ */
  INTVCS(0xfff2, proc_ctrl_c);		/* $FFF2:ERROR ABORT xN^  */
#else
  signal(SIGINT, proc_ctrl_c);
#endif /* __HUMAN68K__ */

  /****************************/
  /* ϐ`FbN         */
  /****************************/
  /* P[XPFϐ SYSROOT `Ăꍇ */
  /*           "$SYSROOT/etc/mkwhatis.db"            */
  p = manpath;
  if (GETENV("SYSROOT", 0, p) >= 0) {
    elimdelm(mkwhatis_db, p);
    strcat2(mkwhatis_db, "\\etc", file_mkwhatis);
  }
  if (FILES(&fbuf, mkwhatis_db, 32) < 0) {
  /* P[XQFA:/etc ɑ݂ꍇ  */
  /*           "A:/etc/mkwhatis.db"   */
    strcpycat(p, "A:\\etc", file_mkwhatis);
    if (FILES(&fbuf, p, 32) >= 0) {
      strcpy(mkwhatis_db, p);
    }
  /* P[XRFmkwhatis ƓfBNgɑ݂ꍇ */
  /*           "`dirname argv[0]`/mkwhatis.db"           */
    else {
      dirname(p, argv[0]);
      elimdelm(mkwhatis_db, p);
      strcat(mkwhatis_db, file_mkwhatis);
      if (FILES(&fbuf, mkwhatis_db, 32) < 0) {
  /* ̑̓G[ */
        FPUTS(nodb, 2); EXIT2(1);
      }
    }
  }
  if (GETENV("MANPATH", 0, manpath) < 0) { strcpy(manpath, Manpath); }
  if (GETENV("MANSEC",  0, mansecs) < 0) { strcpy(mansecs, Mansecs); }
  if (GETENV("UCOMP",   0, ucomp)   < 0) {
#ifdef INTERNAL_GZIP
   ucomp[0] = 0;
#else
   strcpy(ucomp,   Ucomp);
#endif /* INTERNAL_GZIP */
  }
  if (GETENV("temp",    0, temp)    < 0) { strcpy(temp,    Temp);    }
  if (GETENV("AWK",     0, awk)     < 0) { strcpy(awk,     Awk);     }

  /****************************/
  /* `FbN             */
  /****************************/
  argv++;
  while (*argv != NULL)
  {
    if (argv[0][0] != '-')  { break; }
    if (argv[0][1] == '\0') { continue; }
    if (argv[0][2] == '\0') {
      if (argv[0][1] == 'L') { showlic=1; break; } /* show license   */
      if (argv[0][1] == '-') { argv++;    break; } /* no more option */
      if (argv[0][1] == 'M') {
        argv++;
        if (*argv == NULL) { FPUTS(misspath, 2); EXIT2(1); }
        strcpy(manpath, *argv++); continue;
      }
    }
    for (p = &argv[0][1]; *p != 0; p++) {
      if (*p == 'V') { showver=1; continue; } /* show version     */
      if (*p == 'v') { vb = 1;    continue; } /* verbose          */
      /* sȃIvV */
      illegal[ILLPOS] = *p; FPUTS(illegal, 2); showlic=1; break;
    }
    argv++;
  }

  /* license \ */
  if (showlic != 0) {
   FPUTS(version, 2);
#ifdef INTERNAL_GZIP
   FPUTS(gzipcode, 2);
#endif /* INTERNAL_GZIP */
   FPUTS(license, 2);
   FPUTS(usage_mkwhatis, 2);
   EXIT2(1);
  }

  /* o[W\ */
  if (showver != 0) { FPUTS(version, 2); }

  /* mkwhatis.db ǂݍ */
  if (vb != 0) { FPUTS(prsmkwdb, 2); }
  prs_mkwhatisdb(mkwhatis_db);
  if (vb != 0) { FPUTS(done, 2); }

  /* fBNge[u쐬 */
  makedtree();

  /* fBNgc[쐬 */
  if (vb != 0) { FPUTS(mkdtreemes, 2); }
  allocdtree();
  if (vb != 0) { FPUTS(done, 2); }
#ifdef DEBUG
  if (vb != 0) { dumpdtree(); }
#endif /* DEBUG */

  /* mkwhatis.db ǂݍ݁Amkwhatis awk script 𐶐 */
  if (vb != 0) { FPUTS(genmkwhat, 2); }
  gen_mkwhatis();
  if (vb != 0) { FPUTS(done, 2); }

  /* mkwhatis awk script sAwhatis f[^x[X쐬 */
  make_whatis();

  return  0;
}

/*------------------------------------------------------------------*/
/*  gen_mkwhatis(): mkwhatis awk script ֐                    */
/*------------------------------------------------------------------*/
char mkwform1[]="\
BEGIN			{state=0}\x0d\x0a\
NR > 20	{\
  printf(\"%s: This looks like a normal text\\n\",fn)>\"/dev/stderr\";\
  exit\
}\x0d\x0a\
$1 == \"\"		{next}\x0d\x0a\
/^\\.SH/		{sub(\"\\.SH\",\"\")}\x0d\x0a\
/^\\./			{NR--;next}\x0d\x0a\
/\\\\-/			{gsub(\"\\\\\\\\-\",\"-\")}\x0d\x0a\
/.\\b/			{gsub(\".\\b\",\"\")}\x0d\x0a";
char mkwform2[]="	{state=1;next}\x0d\x0a";
char mkwform3[]="	{if(state==2){print\"\";exit}}\x0d\x0a\
state == \"1\"	{\
	n=1;$(NF+1)=\"-\";\
	while($n!=\"-\"){fmt=fmt $n \" \";n++}n++;$NF=\"\";\
	fmt=fmt ss;printf(\"%-22s  -\",fmt);state=2\
}\x0d\x0a\
state == \"2\"	{\
	buf=\"\";\
	while($n!=\"\"){buf=buf \" \" $n;n++}\
	printf(\"%s\",buf);n=1\
}";

static inline void write_mkwhatis(int fno, char * s, int size)
{
  int ret;
  if ((ret = WRITE(fno, s, size)) < 0) { error_in_doscall(ret, NULL); }
}

char mkwformxx[]="$1 == \"";
char mkwformyy[]=" || $1 == \"";
char mkwformzz[]="\"";

void gen_mkwhatis()
{
  char buf1[256], *buf1p = buf1;
  char bufn[256], *bufnp = bufn;
  int fno;

  strcpy(buf1, mkwformxx); buf1p += sizeof(mkwformxx); buf1p--;
  strcpy(bufn, mkwformyy); bufnp += sizeof(mkwformyy); bufnp--;

  elimdelm(awkscript, temp);
  strcat(awkscript, tempname);

  /* AWK XNvgt@CI[v */
  if ((fno=MAKETMP(awkscript,32)) < 0) {error_in_doscall(fno,failtmp);}

  /* tH[Po */
  write_mkwhatis(fno, mkwform1, sizeof(mkwform1)-1);

  /* ʕPo */
  list_scan_id1();
  strcpycat(buf1p, list_get_id1(), mkwformzz);
  write_mkwhatis(fno, buf1, strlen(buf1));
  list_next_id1();

  for ( ;list_more_id1(); list_next_id1()) {
    strcpycat(bufnp, list_get_id1(), mkwformzz);
    write_mkwhatis(fno, bufn, strlen(bufn));
  }

  /* tH[Qo */
  write_mkwhatis(fno, mkwform2, sizeof(mkwform2)-1);

  /* ʕQo */
  list_scan_id2();
  strcpycat(buf1p, list_get_id2(), mkwformzz);
  write_mkwhatis(fno, buf1, strlen(buf1));
  list_next_id2();

  for ( ;list_more_id2(); list_next_id2()) {
    strcpycat(bufnp, list_get_id2(), mkwformzz);
    write_mkwhatis(fno, bufn, strlen(bufn));
  }

  /* tH[Ro */
  write_mkwhatis(fno, mkwform3, sizeof(mkwform3)-1);

  /* AWK XNvgt@CN[Y */
  CLOSE(fno);
}

/*------------------------------------------------------------------*/
/*  make_whatis(): whatis f[^x[X쐬֐                      */
/*------------------------------------------------------------------*/
char awkform1[]="-v ss=(";
char awkform2[]=") -v fn=";
char awkform3[]=" -f ";

void make_whatis()
{
  Directories_t *	Dhd;
  Files_t	*	Fhd;

  char ucompprm[256], awkprm[256];
  char prm[256], *out;
  char path[256];
  char *sec, *name;
  int  ret;

  int  form;
  int  oldindex=0;	/* MANPATH ςǂێ */
  int  fno_o;		/* whatis f[^x[Xpnh */
  int  fno0, fno1;

  elimdelm(tmp1, temp);
  strcat(tmp1, tempname);
  /**************************************/
  /* PATH ɑ݂R}h𓾂  */
  /**************************************/
#ifdef INTERNAL_GZIP
  if (ucomp[0] != 0) {
#endif /* INTERNAL_GZIP */
    if ((ret = EXEC2(2, ucomp, ucompprm, 0)) < 0) {
      strncpy(&failmes[FAILMESP], "UCOMP", 5);
      error_in_doscall(ret, failmes);
    }
#ifdef INTERNAL_GZIP
  }
#endif /* INTERNAL_GZIP */
  if ((ret = EXEC2(2, awk, awkprm, 0)) < 0) {
    strncpy(&failmes[FAILMESP], " AWK ", 5);
    error_in_doscall(ret, failmes);
  }

  /******************/
  /* [vJn     */
  /******************/
  for (Directory__scan(Dhd);		/* fBNǧJn   */
       Directory__more(Dhd);		/* fBNg݂ԁA   */
      )					/* JԂ */
  {
    /* whatis I[v */
    elimdelm(path, Directory__path(Dhd));
    strcat(path, file_whatis);
    if ((fno_o = CREATE(path, 32)) < 0) { error_in_doscall(fno_o, NULL); }

    for ( ;				/*  MANPATH ԂŁA */
         Directory__more(Dhd);		/* fBNg݂ԁA   */
         Directory__next(Dhd))		/* XƃfBNg */
    {
      if (Directory__index(Dhd) != oldindex) {
        oldindex = Directory__index(Dhd);
        break;
      }

      for (form = 1; form >= 0; form--)	/* cat?, man? ̏Ɍ    */
      {
        sec = Directory__section(Dhd);
        formtbl[form][FORM_SEC] = *sec;

        for (File__scan(Dhd, Fhd, form);/* t@ČJn   */
             File__more(Fhd);		/* t@C݂ԁA   */
             File__next(Fhd))		/* Xƃt@C */
        {
          /* man t@C̃t@C𐶐 */
	  elimdelm(prm, Directory__path(Dhd));
	  strcat5(prm, formtbl[form], "\\", File__name(Fhd), dot, sec);
				/* delimitor */
          /* ̎_ł̏óiawk ւ̓́j man t@Cg */
	  out = prm;

     /****************************/
     /* CALL 'UNCOMPESS'         */
     /****************************/
          if (File__compressed(Fhd)) {
            /* óiawk ւ̓́j k man t@CƂȂ */
            out = tmp1;
	    strcat2(prm, dot, File__extension(Fhd));
#ifdef INTERNAL_GZIP
            if (ucomp[0] == 0) { gzipmain(prm, out); }
            else {
#endif /* INTERNAL_GZIP */
	      if ((fno0=   OPEN(prm, 0)) < 0){error_in_doscall(fno0,NULL);}
	      if ((fno1=MAKETMP(out,32)) < 0){error_in_doscall(fno1,failtmp);}
              stdio_exec2(fno0, fno1, ucomp, ucompprm, "UCOMP");
              CLOSE(fno1); CLOSE(fno0);
#ifdef INTERNAL_GZIP
            }
#endif /* INTERNAL_GZIP */
	  }
     /****************************/
     /* CALL 'AWK'               */
     /****************************/
          /* awk ̃p[^ */
          name = File__name(Fhd);
          strcpy(&awkprm[1], awkform1);
          strcat5(&awkprm[1], sec, awkform2, name, awkform3, awkscript);
          awkprm[0] = (unsigned char)strlen(&awkprm[1]);

          if ((fno0=OPEN(out,0)) < 0){error_in_doscall(fno0,failtmp);}
          stdio_exec2(fno0, fno_o, awk, awkprm, " AWK ");
	  CLOSE(fno0);
          if (File__compressed(Fhd)) { DELETE(out); }
        }
      }
    }
    /* whatis N[Y */
    CLOSE(fno_o);
  }
  /* awk XNvg폜 */
  DELETE(awkscript);
}

/*------------------------------------------------------------------*/
/*  error_in_doscall(): DOS CALL G[p֐                   */
/*------------------------------------------------------------------*/
static char * errormessage[] = {
0, 0, 
"w肵t@C͌܂B\x0d\x0a",                 /* 2 */
"w肵fBNg͌܂B\x0d\x0a",               /* 3 */
"I[vĂt@C܂B\x0d\x0a",               /* 4 */
"fBNg{[x̓ANZXł܂B\x0d\x0a", /* 5 */
0, 0,
"ŝɕKvȃ܂B\x0d\x0a",               /* 8 */
0, 0, 0, 0,
"t@C̎wɌ肪܂B\x0d\x0a",                   /*13 */
0, 0, 0, 0, 0,
"̃t@C͏݂ł܂B\x0d\x0a",                 /*19 */
0, 0, 0,
"fBXNςŃt@C܂B\x0d\x0a",           /*23 */
"fBNgςŃt@C܂B\x0d\x0a"        /*24 */
/* 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0 */
};

static char errormessage80[] =
"t@C݂Ă܂B\x0d\x0a";


volatile void error_in_doscall(int code, char * mes)
{
  struct _filbuf fbuf;
  char * mesp;

  code = -code;
  if (code == 80) {
    FPUTS(errormessage80, 2);
  }
  else if (code <= 24) {
    if ((mesp = errormessage[code]) != NULL) { FPUTS(mesp, 2); }
#ifdef DEBUG
    else { fprintf(stderr, "code=(%d)\n", code); }
#endif /* DEBUG */
  }
  if (mes != NULL) { FPUTS(mes, 2); }
  if (FILES(&fbuf, awkscript, 32) >= 0) { DELETE(awkscript); }
  if (FILES(&fbuf, tmp1,      32) >= 0) { DELETE(tmp1); }
  EXIT2(1);
}

#ifdef DEBUG
/*------------------------------------------------------------------*/
/*  dumpdtree(): fobOp֐ifBNgc[\j     */
/*------------------------------------------------------------------*/
void dumpdtree()
{
  Directories_t *	Dhd;
  Files_t	*	Fhd;

  char path[256], *sec;
  int form;

  for (Directory__scan(Dhd);
       Directory__more(Dhd);
       Directory__next(Dhd)) {

    for (form = 1; form >= 0; form--) {

      sec = Directory__section(Dhd);
      formtbl[form][FORM_SEC] = *sec;
      elimdelm(path, Directory__path(Dhd)); strcat(path, formtbl[form]);

      printf("Directory: [%s]\n", path);
    
      for (File__scan(Dhd, Fhd, form);
           File__more(Fhd);
           File__next(Fhd)) {

	if (File__compressed(Fhd) != FALSE) {
          sprintf(path, "%s.%s.%s", File__name(Fhd), sec, File__extension(Fhd));
	}
	else {
          sprintf(path, "%s.%s", File__name(Fhd), sec);
	}
	printf("  filename: [%s]\n", path);
      }
    }
  }
}
#endif /* DEBUG */


/* eof */
