/* parser.c -- parser for `mkwhatis.db'
 * Copyright (C) 1993-1994 Hitoshi Kihara.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef RCSID
static char rcsid[]="$Id: parser.c,v 1.1 1994/10/27 14:07:35 hkl Exp $";
#endif /* RCSID */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __HUMAN68K__
#define __DOS_INLINE__
#include <doslib.h>
#undef _filbuf		/* from <doslib.h> */
#endif /* __HUMAN68K__ */

#include "lexer.h"
#include "parser.h"

/*
 *
 */
#ifdef D_PARSER
char nomem[]="mkwhatis: not enough memory.\x0d\x0a";
#else
extern char nomem[];	/* in mkwhatis.c */
#endif /* D_PARSER */
char noid1[]="mkwhatis: no ID1 found in mkwhatis.db\x0d\x0a";
char noid2[]="mkwhatis: no ID2 found in mkwhatis.db\x0d\x0a";

/*------------------------------------------------------------------*/
/* \̐錾у\bh`                                     */
/*------------------------------------------------------------------*/
typedef struct _id_list_t {
  char * id;
  struct _id_list_t * next;
} ID_List_t;

static ID_List_t id1_top, *id1_handle;
static ID_List_t id2_top, *id2_handle;

/* initializer */
void   list_init()     { id1_top.next = id2_top.next = NULL; }
void   list_add_id(ID_List_t * idlist, char * id)
{
  for (; idlist->next != NULL; idlist = idlist->next);
  idlist = idlist->next = (ID_List_t *)malloc(sizeof(ID_List_t));
  if (idlist == NULL) { FPUTS(nomem,2); EXIT2(1); }
  idlist->id = strdup(id);
  idlist->next = NULL;
}
void   list_add_id1(char * id) { list_add_id(&id1_top, id); }
void   list_add_id2(char * id) { list_add_id(&id2_top, id); }


/* interface functions */
void   list_scan_id1() { id1_handle = id1_top.next; }
int    list_more_id1() { return (int)id1_handle; } /* FALSE if no more */
char * list_get_id1()  { return id1_handle->id; }
void   list_next_id1() { id1_handle = id1_handle->next; }

void   list_scan_id2() { id2_handle = id2_top.next; }
int    list_more_id2() { return (int)id2_handle; } /* FALSE if no more */
char * list_get_id2()  { return id2_handle->id; }
void   list_next_id2() { id2_handle = id2_handle->next; }


/*------------------------------------------------------------------*/
/* \G[                                                   */
/*------------------------------------------------------------------*/
static void prs_sync_token()
{
  for (; lex_lookahead() != NL && lex_lookahead() != DONE; lex_next());
}
static void prs_error_ID1_ID2()
{
  printf("mkwhatis.db(%d): expected ID1 or ID2\n", lex_lineno());
  prs_sync_token();
}
static void prs_error_NL()
{
  printf("mkwhatis.db(%d): expected newline\n", lex_lineno());
  prs_sync_token();
}
static void prs_error_ID()
{
  printf("mkwhatis.db(%d): expected ID\n", lex_lineno());
  prs_sync_token();
}

/*------------------------------------------------------------------*/
/* \͏                                                     */
/*                                                                  */
/*   <mkwhatisdb>       :=      <ID1_declare>   NL                  */
/*                       |      <ID2_declare>   NL                  */
/*   <ID1_declare>      :=      ID1             <ID_list>           */
/*   <ID2_declare>      :=      ID2             <ID_list>           */
/*   <ID_list>          :=      ID              <ID_list>           */
/*                       |                                        */
/*------------------------------------------------------------------*/
static void prs_ID1_declare()
{
  lex_next(); /* parse ID1 */

  if (lex_lookahead() != ID) { prs_error_ID(); return; }

  while (lex_lookahead() == ID) {
    list_add_id1(lex_get_text());
    lex_next();
  }
}
static void prs_ID2_declare()
{
  lex_next(); /* parse ID2 */

  if (lex_lookahead() != ID) { prs_error_ID(); return; }

  while (lex_lookahead() == ID) {
    list_add_id2(lex_get_text());
    lex_next();
  }
}
void prs_mkwhatisdb(char * filename)
{
  int nextchar;

  if (lex_start_scan(filename) == FALSE) {
    fprintf(stderr, "Can't open %s\n",filename);
    exit(1);
  }
  list_init();	/* initialize */
  while ((nextchar = lex_lookahead()) != DONE) {
    switch (nextchar) {
    case NL:
      lex_next(); break;
    case ID1:
      prs_ID1_declare();
      if (lex_lookahead() != NL)  { prs_error_NL(); }
      lex_next(); break;
    case ID2:
      prs_ID2_declare();
      if (lex_lookahead() != NL)  { prs_error_NL(); }
      lex_next(); break;
    default:
      prs_error_ID1_ID2();
    }
  }
  lex_stop_scan();

  /* ʕPŒP͒`Ă邩ǂ */
  list_scan_id1();
  if (list_more_id1() == FALSE) { FPUTS(noid1, 2); EXIT2(1); }

  /* ʕQŒP͒`Ă邩ǂ */
  list_scan_id2();
  if (list_more_id2() == FALSE) { FPUTS(noid2, 2); EXIT2(1); }
}

#ifdef D_PARSER
/*----------------------------------------------*/
/*  main() for debug                            */
/*----------------------------------------------*/
int main(int argc, char **argv)
{
  int ret;
  
  if (argc != 2) {
    fprintf(stderr, "usage: parser file\n");
    exit(1);
  }
  prs_mkwhatisdb(argv[1]);

  FPUTS("\x0d\x0aID1:\x0d\x0a",2);
  for (list_scan_id1(); list_more_id1() != FALSE; list_next_id1()) {
    FPUTS(list_get_id1(), 2); FPUTS("\x0d\x0a",2);
  }
  FPUTS("\x0d\x0aID2:\x0d\x0a",2);
  for (list_scan_id2(); list_more_id2() != FALSE; list_next_id2()) {
    FPUTS(list_get_id2(), 2); FPUTS("\x0d\x0a",2);
  }
}
#endif /* D_PARSER */


/* eof */
